r"""Contain some utility functions to manipulate DataFrames."""

from __future__ import annotations

__all__ = ["drop_duplicates", "generate_vocabulary"]

from arctix.utils.dataframe.removing import drop_duplicates
from arctix.utils.dataframe.vocab import generate_vocabulary



r"""Contain some iterables for paths."""

from __future__ import annotations

__all__ = ["DirFilter", "FileFilter", "PathLister"]

from collections.abc import Iterable, Iterator
from pathlib import Path

from coola.utils import str_indent, str_mapping


class DirFilter(Iterable[Path]):
    r"""Implement an iterable to keep only the directory.

    Args:
        source: The source with the paths.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.utils.iter import DirFilter
    >>> it = DirFilter([Path("tmp/A"), Path("tmp/B"), Path("tmp/C")])
    >>> it
    DirFilter()

    ```
    """

    def __init__(self, source: Iterable[Path]) -> None:
        self._source = source

    def __iter__(self) -> Iterator[Path]:
        for path in self._source:
            if path.is_dir():
                yield path

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def __str__(self) -> str:
        args = str_indent(str_mapping({"source": str_indent(self._source)}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"


class FileFilter(Iterable[Path]):
    r"""Implement an iterable to keep only the files.

    Args:
        source: The source with the paths.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.utils.iter import FileFilter
    >>> it = FileFilter([Path("tmp/A"), Path("tmp/B"), Path("tmp/C")])
    >>> it
    FileFilter()

    ```
    """

    def __init__(self, source: Iterable[Path]) -> None:
        self._source = source

    def __iter__(self) -> Iterator[Path]:
        for path in self._source:
            if path.is_file():
                yield path

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def __str__(self) -> str:
        args = str_indent(str_mapping({"source": str_indent(self._source)}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"


class PathLister(Iterable[Path]):
    r"""Implement an iterable to list paths.

    Args:
        source: The source with the paths.
        pattern: A glob pattern, to return only the matching paths.
        deterministic: If ``True``, the paths are returned in a
            deterministic order.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.utils.iter import PathLister
    >>> it = PathLister([Path("tmp/A"), Path("tmp/B"), Path("tmp/C")])
    >>> it
    PathLister(pattern=*, deterministic=True)

    ```
    """

    def __init__(
        self,
        source: Iterable[Path],
        pattern: str = "*",
        deterministic: bool = True,
    ) -> None:
        self._source = source
        self._pattern = pattern
        self._deterministic = bool(deterministic)

    def __iter__(self) -> Iterator[Path]:
        for path in self._source:
            paths = path.glob(self._pattern)
            if self._deterministic:
                paths = sorted(paths)
            yield from paths

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(pattern={self._pattern}, "
            f"deterministic={self._deterministic})"
        )

    def __str__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "pattern": self._pattern,
                    "deterministic": self._deterministic,
                    "source": str_indent(self._source),
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



r"""Contain code to prepare/preprocess the MultiTHUMOS data.

Information about the MultiTHUMOS dataset can be found in the following
paper:

Every Moment Counts: Dense Detailed Labeling of Actions in Complex
Videos. Yeung S., Russakovsky O., Jin N., Andriluka M., Mori G., Fei-Fei
L. IJCV 2017 (

http://arxiv.org/pdf/1507.05738)

Project page: http://ai.stanford.edu/~syyeung/everymoment.html
"""

from __future__ import annotations

__all__ = [
    "download_data",
    "fetch_data",
    "filter_by_split",
    "generate_split_column",
    "group_by_sequence",
    "is_annotation_path_ready",
    "load_annotation_file",
    "load_data",
    "parse_annotation_lines",
    "prepare_data",
    "to_array_data",
]

import logging
import zipfile
from functools import partial
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING

import numpy as np
import polars as pl
from iden.utils.path import sanitize_path

from arctix.transformer import dataframe as td
from arctix.utils.dataframe import drop_duplicates, generate_vocabulary
from arctix.utils.download import download_url_to_file
from arctix.utils.iter import FileFilter, PathLister
from arctix.utils.mapping import convert_to_dict_of_flat_lists
from arctix.utils.masking import convert_sequences_to_array, generate_mask_from_lengths

if TYPE_CHECKING:
    from collections.abc import Sequence

logger = logging.getLogger(__name__)

ANNOTATION_URL = "http://ai.stanford.edu/~syyeung/resources/multithumos.zip"


class Column:
    ACTION: str = "action"
    ACTION_ID: str = "action_id"
    END_TIME: str = "end_time"
    SEQUENCE_LENGTH: str = "sequence_length"
    SPLIT: str = "split"
    START_TIME: str = "start_time"
    VIDEO: str = "video"
    VIDEO_ID: str = "video_id"


def fetch_data(
    path: Path, remove_duplicate: bool = True, force_download: bool = False
) -> pl.DataFrame:
    r"""Download and load the data for Breakfast dataset.

    Args:
        path: The path where to store the downloaded data.
        remove_duplicate: If ``True``, the duplicate examples are
            removed.
        force_download: If ``True``, the annotations are downloaded
            everytime this function is called. If ``False``,
            the annotations are downloaded only if the
            given path does not contain the annotation data.

    Returns:
        The data in a DataFrame

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.dataset.multithumos import fetch_data
    >>> data = fetch_data(Path("/path/to/data/multithumos/"))  # doctest: +SKIP

    ```
    """
    path = sanitize_path(path)
    download_data(path, force_download)
    return load_data(path, remove_duplicate)


def download_data(path: Path, force_download: bool = False) -> None:
    r"""Download the MultiTHUMOS annotation data.

    Internally, this function downloads the annotations in a temporary
    directory, then extracts the files from the download zip files in
    the temporary directory, and finally moves the extracted files to
    the given path.

    Args:
        path: The path where to store the MultiTHUMOS data.
        force_download: If ``True``, the annotations are downloaded
            everytime this function is called. If ``False``,
            the annotations are downloaded only if the
            given path does not contain the annotation data.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.dataset.multithumos import download_data
    >>> path = Path("/path/to/data")
    >>> download_data(path)  # doctest: +SKIP

    ```
    """
    path = sanitize_path(path)
    if not is_annotation_path_ready(path) or force_download:
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            zip_file = tmp_path.joinpath("multithumos.zip.tmp")
            logger.info(f"downloading MultiTHUMOS annotations data in {zip_file}...")
            download_url_to_file(ANNOTATION_URL, zip_file.as_posix(), progress=True)

            logger.info(f"extracting {zip_file} in {tmp_path}...")
            with zipfile.ZipFile(zip_file, "r") as zip_ref:
                zip_ref.extractall(tmp_path)

            logger.info(f"moving extracted files to {path}...")
            path.mkdir(parents=True, exist_ok=True)
            tmp_path.joinpath("multithumos/README").rename(path.joinpath("README"))
            tmp_path.joinpath("multithumos/class_list.txt").rename(path.joinpath("class_list.txt"))
            tmp_path.joinpath("multithumos/annotations").rename(path.joinpath("annotations"))

    logger.info(f"MultiTHUMOS annotation data are available in {path}")


def is_annotation_path_ready(path: Path) -> bool:
    r"""Indicate if the given path contains the MultiTHUMOS annotation
    data.

    Args:
        path: The path to check.

    Returns:
        ``True`` if the path contains the MultiTHUMOS data,
            otherwise ``False``.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.dataset.multithumos import is_annotation_path_ready
    >>> is_annotation_path_ready(Path("/path/to/data/"))
    False

    ```
    """
    path = sanitize_path(path)
    if not path.joinpath("README").is_file():
        return False
    if not path.joinpath("class_list.txt").is_file():
        return False
    if not path.joinpath("annotations").is_dir():
        return False
    return len(tuple(path.joinpath("annotations").glob("*.txt"))) == 65


def load_data(path: Path, remove_duplicate: bool = True) -> pl.DataFrame:
    r"""Load all the annotations in a DataFrame.

    Args:
        path: The directory where the dataset annotations are stored.
        remove_duplicate: If ``True``, the duplicate rows are removed.

    Returns:
        The annotations in a DataFrame.
    """
    paths = FileFilter(PathLister([sanitize_path(path)], pattern="annotations/*.txt"))
    annotations = list(map(load_annotation_file, paths))
    data = convert_to_dict_of_flat_lists(annotations)
    data = pl.DataFrame(data)
    if remove_duplicate:
        data = drop_duplicates(data)
    if data.select(pl.len()).item():
        data = data.sort(by=[Column.VIDEO, Column.START_TIME])
    return data


def load_annotation_file(path: Path) -> dict[str, list]:
    r"""Load the annotation data from a text file.

    Args:
        path: The file path to the annotation data.

    Returns:
        A dictionary with the action, the start time, and end time
            of each action.
    """
    path = sanitize_path(path)
    if path.suffix != ".txt":
        msg = (
            "Incorrect file extension. This function can only parse `.txt` files "
            f"but received {path.suffix}"
        )
        raise ValueError(msg)
    logger.info(f"Reading {path}...")
    with Path.open(path) as file:
        lines = [x.strip() for x in file.readlines()]

    annotation = parse_annotation_lines(lines)
    annotation[Column.ACTION] = [path.stem] * len(annotation[Column.VIDEO])
    return annotation


def parse_annotation_lines(lines: Sequence[str]) -> dict:
    r"""Parse the action annotation lines and returns a dictionary with
    the prepared data.

    Args:
        lines: The lines to parse.

    Returns:
        A dictionary with the sequence of video names, the start
            time and end time of each action.

    Example usage:

    ```pycon

    >>> from arctix.dataset.multithumos import parse_annotation_lines
    >>> out = parse_annotation_lines(
    ...     [
    ...         "video_validation_0000266 72.80 76.40",
    ...         "video_validation_0000681 44.00 50.90",
    ...         "video_validation_0000682 1.50 5.40",
    ...         "video_validation_0000682 79.30 83.90",
    ...     ]
    ... )
    >>> out
    {'video': ['video_validation_0000266', 'video_validation_0000681', 'video_validation_0000682', 'video_validation_0000682'],
     'start_time': [72.8, 44.0, 1.5, 79.3],
     'end_time': [76.4, 50.9, 5.4, 83.9]}

    ```
    """
    videos = []
    start_time = []
    end_time = []
    for line in (item.strip() for item in lines):
        if not line:
            continue
        video, start, end = line.split(" ")
        videos.append(video)
        start_time.append(float(start))
        end_time.append(float(end))
    return {Column.VIDEO: videos, Column.START_TIME: start_time, Column.END_TIME: end_time}


def prepare_data(frame: pl.DataFrame, split: str = "all") -> tuple[pl.DataFrame, dict]:
    r"""Prepare the data.

    Args:
        frame: The raw DataFrame.
        split: The dataset split. By default, the union of all the
            dataset splits is used.

    Returns:
        A tuple containing the prepared data and the metadata.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.dataset.multithumos import Column, prepare_data
    >>> frame = pl.DataFrame(
    ...     {
    ...         Column.VIDEO: ["video_validation_1", "video_test_2", "video_validation_3", "video_test_4"],
    ...         Column.START_TIME: [72.80, 44.00, 1.50, 17.57],
    ...         Column.END_TIME: [76.40, 50.90, 5.40, 18.33],
    ...         Column.ACTION: ["dribble", "dribble", "dribble", "guard"],
    ...     }
    ... )
    >>> data, metadata = prepare_data(frame)
    >>> data
    shape: (4, 6)
    ┌────────────────────┬────────────┬───────────┬─────────┬───────────┬────────────┐
    │ video              ┆ start_time ┆ end_time  ┆ action  ┆ action_id ┆ split      │
    │ ---                ┆ ---        ┆ ---       ┆ ---     ┆ ---       ┆ ---        │
    │ str                ┆ f32        ┆ f32       ┆ str     ┆ i64       ┆ str        │
    ╞════════════════════╪════════════╪═══════════╪═════════╪═══════════╪════════════╡
    │ video_test_2       ┆ 44.0       ┆ 50.900002 ┆ dribble ┆ 0         ┆ test       │
    │ video_test_4       ┆ 17.57      ┆ 18.33     ┆ guard   ┆ 1         ┆ test       │
    │ video_validation_1 ┆ 72.800003  ┆ 76.400002 ┆ dribble ┆ 0         ┆ validation │
    │ video_validation_3 ┆ 1.5        ┆ 5.4       ┆ dribble ┆ 0         ┆ validation │
    └────────────────────┴────────────┴───────────┴─────────┴───────────┴────────────┘
    >>> metadata
    {'vocab_action': Vocabulary(
      counter=Counter({'dribble': 3, 'guard': 1}),
      index_to_token=('dribble', 'guard'),
      token_to_index={'dribble': 0, 'guard': 1},
    )}

    ```
    """
    vocab_action = generate_vocabulary(frame, col=Column.ACTION).sort_by_count()
    transformer = td.Sequential(
        [
            td.Sort(columns=[Column.VIDEO, Column.START_TIME]),
            td.Cast(columns=[Column.START_TIME, Column.END_TIME], dtype=pl.Float32),
            td.StripChars(columns=[Column.ACTION, Column.VIDEO]),
            td.TokenToIndex(
                vocab=vocab_action, token_column=Column.ACTION, index_column=Column.ACTION_ID
            ),
            td.Function(generate_split_column),
            td.Function(partial(filter_by_split, split=split)),
        ]
    )
    out = transformer.transform(frame)
    return out, {"vocab_action": vocab_action}


def generate_split_column(frame: pl.DataFrame) -> pl.DataFrame:
    r"""Generate the split column from the video name column.

    Args:
        frame: The input DataFrame with the video name column.

    Returns:
        The output DataFrame with the additional split column.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.dataset.multithumos import Column, generate_split_column
    >>> frame = pl.DataFrame(
    ...     {
    ...         Column.VIDEO: ["video_validation_1", "video_test_2", "video_validation_3", "video_test_4"],
    ...         Column.ACTION_ID: [0, 2, 5, 1],
    ...     }
    ... )
    >>> out = generate_split_column(frame)
    >>> out
    shape: (4, 3)
    ┌────────────────────┬───────────┬────────────┐
    │ video              ┆ action_id ┆ split      │
    │ ---                ┆ ---       ┆ ---        │
    │ str                ┆ i64       ┆ str        │
    ╞════════════════════╪═══════════╪════════════╡
    │ video_validation_1 ┆ 0         ┆ validation │
    │ video_test_2       ┆ 2         ┆ test       │
    │ video_validation_3 ┆ 5         ┆ validation │
    │ video_test_4       ┆ 1         ┆ test       │
    └────────────────────┴───────────┴────────────┘

    ```
    """
    return frame.with_columns(
        pl.col(Column.VIDEO).str.split_exact(by="_", n=3).struct[1].alias(Column.SPLIT)
    )


def filter_by_split(frame: pl.DataFrame, split: str = "all") -> pl.DataFrame:
    r"""Filter the DataFrame to keep only the rows associated to a
    dataset split.

    Args:
        frame: The DataFrame to filter.
        split: The dataset split. By default, the union of all the
            dataset splits is used.

    Returns:
        The filtered DataFrame.
    """
    splits = {split}
    if split == "all":
        splits = {"validation", "test"}
    return frame.filter(pl.col(Column.SPLIT).is_in(splits))


def group_by_sequence(frame: pl.DataFrame) -> pl.DataFrame:
    r"""Group the DataFrame by sequences of actions.

    Args:
        frame: The input DataFrame.

    Returns:
        The DataFrame after the grouping.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.dataset.multithumos import Column, group_by_sequence
    >>> frame = pl.DataFrame(
    ...     {
    ...         Column.VIDEO: [
    ...             "video_validation_1",
    ...             "video_validation_1",
    ...             "video_validation_1",
    ...             "video_validation_2",
    ...             "video_validation_2",
    ...             "video_validation_2",
    ...             "video_validation_2",
    ...         ],
    ...         Column.START_TIME: [1.50, 17.57, 79.30, 2.97, 4.54, 20.22, 27.42],
    ...         Column.END_TIME: [5.40, 18.33, 83.90, 3.60, 5.07, 20.49, 30.23],
    ...         Column.ACTION: [
    ...             "dribble",
    ...             "guard",
    ...             "dribble",
    ...             "guard",
    ...             "guard",
    ...             "guard",
    ...             "shoot",
    ...         ],
    ...         Column.ACTION_ID: [1, 0, 1, 0, 0, 0, 2],
    ...         Column.SPLIT: [
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...         ],
    ...     },
    ...     schema={
    ...         Column.VIDEO: pl.String,
    ...         Column.START_TIME: pl.Float32,
    ...         Column.END_TIME: pl.Float32,
    ...         Column.ACTION: pl.String,
    ...         Column.ACTION_ID: pl.Int64,
    ...         Column.SPLIT: pl.String,
    ...     },
    ... )
    >>> groups = group_by_sequence(frame)
    >>> groups
    shape: (2, 6)
    ┌─────────────────┬────────────┬─────────────┬─────────────────┬─────────────────┬─────────────────┐
    │ video           ┆ split      ┆ action_id   ┆ start_time      ┆ end_time        ┆ sequence_length │
    │ ---             ┆ ---        ┆ ---         ┆ ---             ┆ ---             ┆ ---             │
    │ str             ┆ str        ┆ list[i64]   ┆ list[f32]       ┆ list[f32]       ┆ u32             │
    ╞═════════════════╪════════════╪═════════════╪═════════════════╪═════════════════╪═════════════════╡
    │ video_validatio ┆ validation ┆ [1, 0, 1]   ┆ [1.5, 17.57,    ┆ [5.4, 18.33,    ┆ 3               │
    │ n_1             ┆            ┆             ┆ 79.300003]      ┆ 83.900002]      ┆                 │
    │ video_validatio ┆ validation ┆ [0, 0, … 2] ┆ [2.97, 4.54, …  ┆ [3.6, 5.07, …   ┆ 4               │
    │ n_2             ┆            ┆             ┆ 27.42]          ┆ 30.23]          ┆                 │
    └─────────────────┴────────────┴─────────────┴─────────────────┴─────────────────┴─────────────────┘

    ```
    """
    return (
        frame.group_by([Column.VIDEO])
        .agg(
            pl.first(Column.SPLIT),
            pl.col(Column.ACTION_ID).alias(Column.ACTION_ID),
            pl.col(Column.START_TIME).alias(Column.START_TIME),
            pl.col(Column.END_TIME).alias(Column.END_TIME),
            pl.len().alias(Column.SEQUENCE_LENGTH),
        )
        .sort(by=[Column.VIDEO])
    )


def to_array_data(frame: pl.DataFrame) -> dict[str, np.ndarray]:
    r"""Convert a DataFrame to a dictionary of arrays.

    Args:
        frame: The input DataFrame.

    Returns:
        The dictionary of arrays.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.dataset.multithumos import Column, to_array_data
    >>> frame = pl.DataFrame(
    ...     {
    ...         Column.VIDEO: [
    ...             "video_validation_1",
    ...             "video_validation_1",
    ...             "video_validation_1",
    ...             "video_validation_2",
    ...             "video_validation_2",
    ...             "video_validation_2",
    ...             "video_validation_2",
    ...         ],
    ...         Column.START_TIME: [1.0, 17.0, 79.0, 2.0, 4.0, 20.0, 27.0],
    ...         Column.END_TIME: [5.0, 18.0, 83.0, 3.0, 5.0, 20.0, 30.0],
    ...         Column.ACTION: [
    ...             "dribble",
    ...             "guard",
    ...             "dribble",
    ...             "guard",
    ...             "guard",
    ...             "guard",
    ...             "shoot",
    ...         ],
    ...         Column.ACTION_ID: [1, 0, 1, 0, 0, 0, 2],
    ...         Column.SPLIT: [
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...             "validation",
    ...         ],
    ...     },
    ...     schema={
    ...         Column.VIDEO: pl.String,
    ...         Column.START_TIME: pl.Float32,
    ...         Column.END_TIME: pl.Float32,
    ...         Column.ACTION: pl.String,
    ...         Column.ACTION_ID: pl.Int64,
    ...         Column.SPLIT: pl.String,
    ...     },
    ... )
    >>> arrays = to_array_data(frame)
    >>> arrays
    {'sequence_length': array([3, 4]),
     'split': array(['validation', 'validation'], dtype='<U10'),
     'action_id': masked_array(
      data=[[1, 0, 1, --],
            [0, 0, 0, 2]],
      mask=[[False, False, False,  True],
            [False, False, False, False]],
      fill_value=999999),
     'start_time': masked_array(
      data=[[1.0, 17.0, 79.0, --],
            [2.0, 4.0, 20.0, 27.0]],
      mask=[[False, False, False,  True],
            [False, False, False, False]],
      fill_value=1e+20),
     'end_time': masked_array(
      data=[[5.0, 18.0, 83.0, --],
            [3.0, 5.0, 20.0, 30.0]],
      mask=[[False, False, False,  True],
            [False, False, False, False]],
      fill_value=1e+20)}

    ```
    """
    groups = group_by_sequence(frame)
    lengths = groups.get_column(Column.SEQUENCE_LENGTH).to_numpy()
    mask = generate_mask_from_lengths(lengths)
    return {
        Column.SEQUENCE_LENGTH: lengths.astype(int),
        Column.SPLIT: groups.get_column(Column.SPLIT).to_numpy().astype(str),
        Column.ACTION_ID: np.ma.masked_array(
            data=convert_sequences_to_array(
                groups.get_column(Column.ACTION_ID).to_list(), dtype=int, max_len=mask.shape[1]
            ).astype(int),
            mask=mask,
        ),
        Column.START_TIME: np.ma.masked_array(
            data=convert_sequences_to_array(
                groups.get_column(Column.START_TIME).to_list(), dtype=float, max_len=mask.shape[1]
            ).astype(float),
            mask=mask,
        ),
        Column.END_TIME: np.ma.masked_array(
            data=convert_sequences_to_array(
                groups.get_column(Column.END_TIME).to_list(), dtype=float, max_len=mask.shape[1]
            ).astype(float),
            mask=mask,
        ),
    }


if __name__ == "__main__":  # pragma: no cover
    import os

    logging.basicConfig(level=logging.DEBUG)

    path = Path(os.environ["ARCTIX_DATA_PATH"]).joinpath("multithumos")
    raw_data = fetch_data(path)
    data, metadata = prepare_data(raw_data)
    logger.info(f"data:\n{data}")
    logger.info(f"metadata:\n{metadata}")



r"""Contain code to download and prepare the Breakfast data.

Information about the Breakfast dataset can be found in the following
paper:

The Language of Actions: Recovering the Syntax and Semantics of Goal-
Directed Human Activities. Kuehne, Arslan, and Serre. CVPR 2014.

Project page:

https://serre-lab.clps.brown.edu/resource/breakfast-actions-dataset/

Data can be downloaded at

https://serre-lab.clps.brown.edu/resource/breakfast-actions-dataset/#Downloads

The documentation assumes the data are downloaded in the directory `/path/to/data/breakfast/`.
"""

from __future__ import annotations

__all__ = [
    "download_data",
    "fetch_data",
    "filter_by_split",
    "group_by_sequence",
    "load_annotation_file",
    "load_data",
    "parse_annotation_lines",
    "prepare_data",
    "to_array_data",
]

import logging
import tarfile
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import polars as pl
from iden.utils.path import sanitize_path

from arctix.transformer import dataframe as td
from arctix.utils.dataframe import drop_duplicates, generate_vocabulary
from arctix.utils.download import download_drive_file
from arctix.utils.iter import FileFilter, PathLister
from arctix.utils.mapping import convert_to_dict_of_flat_lists
from arctix.utils.masking import convert_sequences_to_array, generate_mask_from_lengths

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)

URLS = {
    "segmentation_coarse": "https://drive.google.com/open?id=1R3z_CkO1uIOhu4y2Nh0pCHjQQ2l-Ab9E",
    "segmentation_fine": "https://drive.google.com/open?id=1Alg_xjefEFOOpO_6_RnelWiNqbJlKhVF",
}
COOKING_ACTIVITIES = (
    "cereals",
    "coffee",
    "friedegg",
    "juice",
    "milk",
    "pancake",
    "salat",
    "sandwich",
    "scrambledegg",
    "tea",
)
NUM_COOKING_ACTIVITIES = {
    "cereals": 214,
    "coffee": 100,
    "friedegg": 198,
    "juice": 187,
    "milk": 224,
    "pancake": 173,
    "salat": 185,
    "sandwich": 197,
    "scrambledegg": 188,
    "tea": 223,
}

PART1 = tuple(f"P{i:02d}" for i in range(3, 16))
PART2 = tuple(f"P{i:02d}" for i in range(16, 29))
PART3 = tuple(f"P{i:02d}" for i in range(29, 42))
PART4 = tuple(f"P{i:02d}" for i in range(42, 55))

DATASET_SPLITS = {
    "all": sorted(PART1 + PART2 + PART3 + PART4),
    "minitrain1": sorted(PART2 + PART3),
    "minitrain2": sorted(PART3 + PART4),
    "minitrain3": sorted(PART1 + PART4),
    "minitrain4": sorted(PART1 + PART2),
    "minival1": sorted(PART4),
    "minival2": sorted(PART1),
    "minival3": sorted(PART2),
    "minival4": sorted(PART3),
    "test1": sorted(PART1),
    "test2": sorted(PART2),
    "test3": sorted(PART3),
    "test4": sorted(PART4),
    "train1": sorted(PART2 + PART3 + PART4),
    "train2": sorted(PART1 + PART3 + PART4),
    "train3": sorted(PART1 + PART2 + PART4),
    "train4": sorted(PART1 + PART2 + PART3),
}


class Column:
    ACTION: str = "action"
    ACTION_ID: str = "action_id"
    COOKING_ACTIVITY: str = "cooking_activity"
    COOKING_ACTIVITY_ID: str = "cooking_activity_id"
    END_TIME: str = "end_time"
    PERSON: str = "person"
    PERSON_ID: str = "person_id"
    START_TIME: str = "start_time"
    SEQUENCE_LENGTH: str = "sequence_length"


def fetch_data(
    path: Path, name: str, remove_duplicate: bool = True, force_download: bool = False
) -> pl.DataFrame:
    r"""Download and load the data for Breakfast dataset.

    Args:
        path: The path where to store the downloaded data.
        name: The name of the dataset. The valid names are
            ``'segmentation_coarse'`` and ``'segmentation_fine'``.
        remove_duplicate: If ``True``, the duplicate examples are
            removed.
        force_download: If ``True``, the annotations are downloaded
            everytime this function is called. If ``False``,
            the annotations are downloaded only if the
            given path does not contain the annotation data.

    Returns:
        The data in a DataFrame

    Raises:
        RuntimeError: if the name is incorrect

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.dataset.breakfast import fetch_data
    >>> data = fetch_data(
    ...     Path("/path/to/data/breakfast/"), "segmentation_coarse"
    ... )  # doctest: +SKIP

    ```
    """
    if name not in (valid_names := set(URLS.keys())):
        msg = f"Incorrect name: {name}. Valid names are: {valid_names}"
        raise RuntimeError(msg)
    path = sanitize_path(path)
    download_data(path, force_download)
    return load_data(path.joinpath(name), remove_duplicate)


def download_data(path: Path, force_download: bool = False) -> None:
    r"""Download the Breakfast annotations.

    Args:
        path: The path where to store the downloaded data.
        force_download: If ``True``, the annotations are downloaded
            everytime this function is called. If ``False``,
            the annotations are downloaded only if the
            given path does not contain the annotation data.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.dataset.breakfast import download_data
    >>> download_data(Path("/path/to/data/breakfast/"))  # doctest: +SKIP

    ```
    """
    path = sanitize_path(path)
    logger.info(f"Downloading Breakfast dataset annotations in {path}...")
    for name, url in URLS.items():
        if not path.joinpath(name).is_dir() or force_download:
            tar_file = path.joinpath(f"{name}.tar.gz")
            download_drive_file(url, tar_file, quiet=False, fuzzy=True)
            tarfile.open(tar_file).extractall(path)  # noqa: S202
            tar_file.unlink(missing_ok=True)


def load_data(path: Path, remove_duplicate: bool = True) -> pl.DataFrame:
    r"""Load all the annotations in a DataFrame.

    Args:
        path: The directory where the dataset annotations are stored.
        remove_duplicate: If ``True``, the duplicate rows are removed.

    Returns:
        The annotations in a DataFrame.
    """
    paths = FileFilter(PathLister([sanitize_path(path)], pattern="**/*.txt"))
    annotations = list(map(load_annotation_file, paths))
    data = convert_to_dict_of_flat_lists(annotations)
    data = pl.DataFrame(data)
    if remove_duplicate:
        data = drop_duplicates(data)
    if data.select(pl.len()).item():
        data = data.sort(by=[Column.COOKING_ACTIVITY, Column.PERSON, Column.START_TIME])
    return data


def load_annotation_file(path: Path) -> dict[str, list]:
    r"""Load the annotation data from a text file.

    Args:
        path: The file path to the annotation data.

    Returns:
        A dictionary with the action, the start time, and end time
            of each action.
    """
    path = sanitize_path(path)
    if path.suffix != ".txt":
        msg = (
            "Incorrect file extension. This function can only parse `.txt` files "
            f"but received {path.suffix}"
        )
        raise ValueError(msg)
    logger.info(f"Reading {path}...")
    with Path.open(path) as file:
        lines = [x.strip() for x in file.readlines()]

    annotation = parse_annotation_lines(lines)
    person_id = path.stem.split("_", maxsplit=1)[0]
    cooking_activity = path.stem.rsplit("_", maxsplit=1)[-1]
    annotation[Column.PERSON] = [person_id] * len(lines)
    annotation[Column.COOKING_ACTIVITY] = [cooking_activity] * len(lines)
    return annotation


def parse_annotation_lines(lines: Sequence[str]) -> dict:
    r"""Parse the action annotation lines and returns a dictionary with
    the prepared data.

    Args:
        lines: The lines to parse.

    Returns:
        A dictionary with the sequence of actions, the start
            time and end time of each action.
    """
    actions = []
    start_time = []
    end_time = []
    for line in lines:
        pair_time, action = line.strip().split()
        actions.append(action)
        start, end = pair_time.split("-")
        start_time.append(float(start))
        end_time.append(float(end))
    return {Column.ACTION: actions, Column.START_TIME: start_time, Column.END_TIME: end_time}


def filter_by_split(frame: pl.DataFrame, split: str = "all") -> pl.DataFrame:
    r"""Filter the DataFrame to keep only the rows associated to a
    dataset split.

    Args:
        frame: The DataFrame to filter.
        split: The dataset split. By default, the union of all the
            dataset splits is used.

    Returns:
        The filtered DataFrame.
    """
    persons = DATASET_SPLITS[split]
    return frame.filter(pl.col(Column.PERSON).is_in(persons))


def prepare_data(frame: pl.DataFrame, split: str = "all") -> tuple[pl.DataFrame, dict]:
    r"""Prepare the data.

    Args:
        frame: The raw DataFrame.
        split: The dataset split. By default, the union of all the
            dataset splits is used.

    Returns:
        A tuple containing the prepared data and the metadata.
    """
    vocab_action = generate_vocabulary(frame, col=Column.ACTION).sort_by_count()
    vocab_person = generate_vocabulary(frame, col=Column.PERSON).sort_by_count()
    vocab_activity = (
        generate_vocabulary(frame, col=Column.COOKING_ACTIVITY).sort_by_token().sort_by_count()
    )
    transformer = td.Sequential(
        [
            td.Sort(columns=[Column.COOKING_ACTIVITY, Column.PERSON, Column.START_TIME]),
            td.Cast(columns=[Column.START_TIME, Column.END_TIME], dtype=pl.Float32),
            td.StripChars(columns=[Column.ACTION, Column.PERSON, Column.COOKING_ACTIVITY]),
            td.Function(partial(filter_by_split, split=split)),
            td.TokenToIndex(
                vocab=vocab_action, token_column=Column.ACTION, index_column=Column.ACTION_ID
            ),
            td.TokenToIndex(
                vocab=vocab_person, token_column=Column.PERSON, index_column=Column.PERSON_ID
            ),
            td.TokenToIndex(
                vocab=vocab_activity,
                token_column=Column.COOKING_ACTIVITY,
                index_column=Column.COOKING_ACTIVITY_ID,
            ),
        ]
    )
    out = transformer.transform(frame)
    return out, {
        "vocab_action": vocab_action,
        "vocab_activity": vocab_activity,
        "vocab_person": vocab_person,
    }


def group_by_sequence(frame: pl.DataFrame) -> pl.DataFrame:
    r"""Group the DataFrame by sequences of actions.

    Args:
        frame: The input DataFrame.

    Returns:
        The DataFrame after the grouping.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.dataset.breakfast import Column, group_by_sequence
    >>> frame = pl.DataFrame(
    ...     {
    ...         Column.START_TIME: [1.0, 31.0, 151.0, 429.0, 576.0, 706.0, 1.0, 48.0, 216.0, 566.0],
    ...         Column.END_TIME: [30.0, 150.0, 428.0, 575.0, 705.0, 836.0, 47.0, 215.0, 565.0, 747.0],
    ...         Column.ACTION_ID: [0, 2, 5, 1, 3, 0, 0, 1, 4, 0],
    ...         Column.PERSON_ID: [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
    ...         Column.COOKING_ACTIVITY_ID: [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
    ...     }
    ... )
    >>> groups = group_by_sequence(frame)
    >>> groups
    shape: (2, 6)
    ┌───────────┬──────────────────┬─────────────┬─────────────────┬─────────────────┬─────────────────┐
    │ person_id ┆ cooking_activity ┆ action_id   ┆ start_time      ┆ end_time        ┆ sequence_length │
    │ ---       ┆ _id              ┆ ---         ┆ ---             ┆ ---             ┆ ---             │
    │ i64       ┆ ---              ┆ list[i64]   ┆ list[f64]       ┆ list[f64]       ┆ u32             │
    │           ┆ i64              ┆             ┆                 ┆                 ┆                 │
    ╞═══════════╪══════════════════╪═════════════╪═════════════════╪═════════════════╪═════════════════╡
    │ 0         ┆ 0                ┆ [0, 2, … 0] ┆ [1.0, 31.0, …   ┆ [30.0, 150.0, … ┆ 6               │
    │           ┆                  ┆             ┆ 706.0]          ┆ 836.0]          ┆                 │
    │ 1         ┆ 1                ┆ [0, 1, … 0] ┆ [1.0, 48.0, …   ┆ [47.0, 215.0, … ┆ 4               │
    │           ┆                  ┆             ┆ 566.0]          ┆ 747.0]          ┆                 │
    └───────────┴──────────────────┴─────────────┴─────────────────┴─────────────────┴─────────────────┘

    ```
    """
    return (
        frame.group_by([Column.PERSON_ID, Column.COOKING_ACTIVITY_ID])
        .agg(
            pl.col(Column.ACTION_ID).alias(Column.ACTION_ID),
            pl.col(Column.START_TIME).alias(Column.START_TIME),
            pl.col(Column.END_TIME).alias(Column.END_TIME),
            pl.len().alias(Column.SEQUENCE_LENGTH),
        )
        .sort(by=[Column.PERSON_ID, Column.COOKING_ACTIVITY_ID])
    )


def to_array_data(frame: pl.DataFrame) -> dict[str, np.ndarray]:
    r"""Convert a DataFrame to a dictionary of arrays.

    Args:
        frame: The input DataFrame.

    Returns:
        The dictionary of arrays.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.dataset.breakfast import Column, to_array_data
    >>> frame = pl.DataFrame(
    ...     {
    ...         Column.START_TIME: [1.0, 31.0, 151.0, 429.0, 576.0, 706.0, 1.0, 48.0, 216.0, 566.0],
    ...         Column.END_TIME: [30.0, 150.0, 428.0, 575.0, 705.0, 836.0, 47.0, 215.0, 565.0, 747.0],
    ...         Column.ACTION_ID: [0, 2, 5, 1, 3, 0, 0, 1, 4, 0],
    ...         Column.PERSON_ID: [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
    ...         Column.COOKING_ACTIVITY_ID: [0, 0, 0, 0, 0, 0, 1, 1, 1, 1],
    ...     }
    ... )
    >>> arrays = to_array_data(frame)
    >>> arrays
    {'sequence_length': array([6, 4]), 'person_id': array([0, 1]),
     'cooking_activity_id': array([0, 1]),
     'action_id': masked_array(
      data=[[0, 2, 5, 1, 3, 0],
            [0, 1, 4, 0, --, --]],
      mask=[[False, False, False, False, False, False],
            [False, False, False, False,  True,  True]],
      fill_value=999999),
     'start_time': masked_array(
      data=[[1.0, 31.0, 151.0, 429.0, 576.0, 706.0],
            [1.0, 48.0, 216.0, 566.0, --, --]],
      mask=[[False, False, False, False, False, False],
            [False, False, False, False,  True,  True]],
      fill_value=1e+20),
     'end_time': masked_array(
      data=[[30.0, 150.0, 428.0, 575.0, 705.0, 836.0],
            [47.0, 215.0, 565.0, 747.0, --, --]],
      mask=[[False, False, False, False, False, False],
            [False, False, False, False,  True,  True]],
      fill_value=1e+20)}

    ```
    """
    groups = group_by_sequence(frame)
    lengths = groups.get_column(Column.SEQUENCE_LENGTH).to_numpy()
    mask = generate_mask_from_lengths(lengths)
    return {
        Column.SEQUENCE_LENGTH: lengths.astype(int),
        Column.PERSON_ID: groups.get_column(Column.PERSON_ID).to_numpy().astype(int),
        Column.COOKING_ACTIVITY_ID: groups.get_column(Column.COOKING_ACTIVITY_ID)
        .to_numpy()
        .astype(int),
        Column.ACTION_ID: np.ma.masked_array(
            data=convert_sequences_to_array(
                groups.get_column(Column.ACTION_ID).to_list(), max_len=mask.shape[1]
            ).astype(int),
            mask=mask,
        ),
        Column.START_TIME: np.ma.masked_array(
            data=convert_sequences_to_array(
                groups.get_column(Column.START_TIME).to_list(), max_len=mask.shape[1]
            ).astype(float),
            mask=mask,
        ),
        Column.END_TIME: np.ma.masked_array(
            data=convert_sequences_to_array(
                groups.get_column(Column.END_TIME).to_list(), max_len=mask.shape[1]
            ).astype(float),
            mask=mask,
        ),
    }


if __name__ == "__main__":  # pragma: no cover
    import os

    logging.basicConfig(level=logging.DEBUG)

    path = Path(os.environ["ARCTIX_DATA_PATH"]).joinpath("breakfast")
    download_data(path)
    data_raw = fetch_data(path, name="segmentation_coarse")
    data, metadata = prepare_data(data_raw)
    logger.info(f"data:\n{data}")



r"""Define some PyTest fixtures."""

from __future__ import annotations

__all__ = ["gdown_available", "requests_available", "tqdm_available"]

import pytest

from arctix.utils.imports import (
    is_gdown_available,
    is_requests_available,
    is_tqdm_available,
)

gdown_available = pytest.mark.skipif(not is_gdown_available(), reason="Require gdown")
requests_available = pytest.mark.skipif(not is_requests_available(), reason="Require requests")
tqdm_available = pytest.mark.skipif(not is_tqdm_available(), reason="Require tqdm")



r"""Contain utility functions to generate or manipulate masks."""

from __future__ import annotations

__all__ = ["convert_sequences_to_array", "generate_mask_from_lengths"]

from typing import TYPE_CHECKING, Any

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import DTypeLike


def convert_sequences_to_array(
    data: Sequence[Sequence], max_len: int, dtype: DTypeLike = None, padded_value: Any = 0
) -> np.ndarray:
    r"""Convert a list of sequences to a ``numpy.ndarray``.

    The sequences can have different lengths, so the sequences are
    padded to all have the same length.

    Args:
        data: The list of sequence to convert to an array.
        max_len: The maximum sequence length which is used to define
            the second dimension of the array. If a sequence is longer,
            it is truncated to the maximum sequence length.
        dtype: The data type of the generated array.
        padded_value: The value used to pad the sequences.

    Returns:
        The generated array.

    Example usage:

    ```pycon

    >>> from arctix.utils.masking import convert_sequences_to_array
    >>> arr = convert_sequences_to_array([[1, 2, 3], [9, 8, 7, 6, 5], [1]], max_len=5)
    >>> arr
    array([[1, 2, 3, 0, 0],
           [9, 8, 7, 6, 5],
           [1, 0, 0, 0, 0]])
    >>> arr = convert_sequences_to_array(
    ...     [[1, 2, 3], [9, 8, 7, 6, 5], [1]],
    ...     max_len=5,
    ...     dtype=float,
    ...     padded_value=-1,
    ... )
    >>> arr
    array([[ 1.,  2.,  3., -1., -1.],
           [ 9.,  8.,  7.,  6.,  5.],
           [ 1., -1., -1., -1., -1.]])

    ```
    """
    array = np.full((len(data), max_len), fill_value=padded_value, dtype=dtype)
    for i, values in enumerate(data):
        length = min(len(values), max_len)
        array[i, :length] = values[:length]
    return array


def generate_mask_from_lengths(lengths: np.ndarray, max_len: int | None = None) -> np.ndarray:
    r"""Generate a mask from the sequences lengths.

    The mask indicates masked values. ``True`` indicates a
    masked/invalid value and ``False`` indicates a valid value.

    Args:
        lengths: The lengths of each sequence in the batch.
        max_len: The maximum sequence length. If ``None``, the maximum
            length is computed based on the given lengths.

    Returns:
        The generated mask of shape ``(batch_size, max_len)``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arctix.utils.masking import generate_mask_from_lengths
    >>> mask = generate_mask_from_lengths(lengths=np.array([4, 3, 5, 3, 2]))
    >>> mask
    array([[False, False, False, False,  True],
           [False, False, False,  True,  True],
           [False, False, False, False, False],
           [False, False, False,  True,  True],
           [False, False,  True,  True,  True]])

    ```
    """
    batch_size = lengths.shape[0]
    if max_len is None:
        max_len = lengths.max(initial=0)
    lengths = np.broadcast_to(lengths.reshape(batch_size, 1), (batch_size, max_len))
    indices = np.broadcast_to(np.arange(max_len).reshape(1, max_len), (batch_size, max_len))
    return indices >= lengths



r"""Contain some utility functions to manipulate mappings."""

from __future__ import annotations

__all__ = ["convert_to_dict_of_flat_lists"]


from typing import TYPE_CHECKING

from coola.nested import convert_to_dict_of_lists

if TYPE_CHECKING:
    from collections.abc import Hashable


def convert_to_dict_of_flat_lists(
    seq_of_mappings: list[dict[Hashable, list]]
) -> dict[Hashable, list]:
    r"""Convert a sequence of mappings to a dictionary of lists.

    All the dictionaries should have the same keys. The first
    mapping in the sequence is used to find the keys.
    The lists of lists are converted to flat lists.

    Args:
        seq_of_mappings: The sequence of mappings to convert.

    Returns:
        A dictionary of lists.

    Example usage:

    ```pycon

    >>> from arctix.utils.mapping import convert_to_dict_of_flat_lists
    >>> convert_to_dict_of_flat_lists(
    ...     [
    ...         {"key1": [1, 2], "key2": [10, 11]},
    ...         {"key1": [2], "key2": [20]},
    ...         {"key1": [3, 4, 5], "key2": [30, 31, 32]},
    ...     ]
    ... )
    {'key1': [1, 2, 2, 3, 4, 5], 'key2': [10, 11, 20, 30, 31, 32]}

    ```
    """
    mapping = convert_to_dict_of_lists(seq_of_mappings)
    return {key: [v for value in values for v in value] for key, values in mapping.items()}



r"""Contain ``polars.DataFrame`` transformers to convert some columns to
a new data type."""

from __future__ import annotations

__all__ = ["CastDataFrameTransformer"]

from typing import TYPE_CHECKING, Any

import polars as pl

from arctix.transformer.dataframe.base import BaseDataFrameTransformer
from arctix.utils.imports import is_tqdm_available

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

if is_tqdm_available():
    from tqdm import tqdm
else:  # pragma: no cover

    def tqdm(it: Iterable, *args: Any, **kwargs: Any) -> Iterable:  # noqa: ARG001
        return it


class CastDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a transformer to convert some columns to a new data
    type.

    Args:
        columns: The columns to convert.
        dtype: The target data type.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.transformer.dataframe import Cast
    >>> transformer = Cast(columns=["col1", "col3"], dtype=pl.Int32)
    >>> transformer
    CastDataFrameTransformer(columns=('col1', 'col3'), dtype=Int32)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ str  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i32  ┆ str  ┆ i32  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    def __init__(self, columns: Sequence[str], dtype: type[pl.DataType]) -> None:
        self._columns = tuple(columns)
        self._dtype = dtype

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(columns={self._columns}, dtype={self._dtype})"

    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        for col in tqdm(self._columns, desc=f"converting to {self._dtype}"):
            frame = frame.with_columns(frame.select(pl.col(col).cast(self._dtype)))
        return frame



r"""Contain utility functions for DataFrame and vocabulary."""

from __future__ import annotations

__all__ = ["generate_vocabulary"]

from collections import Counter
from typing import TYPE_CHECKING

from arctix.utils.vocab import Vocabulary

if TYPE_CHECKING:
    import polars as pl


def generate_vocabulary(frame: pl.DataFrame, col: str) -> Vocabulary:
    r"""Compute a vocabulary based on the content of a column.

    Args:
        frame: The input DataFrame.
        col: The name of the column used to compute the vocabulary.

    Returns:
        The generated vocabulary.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> import polars as pl
    >>> from arctix.utils.dataframe import generate_vocabulary
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a", "b", "c", "d", "a"],
    ...     }
    ... )
    >>> vocab1 = generate_vocabulary(frame, col="col1")
    >>> vocab1
    Vocabulary(
      counter=Counter({1: 1, 2: 1, 3: 1, 4: 1, 5: 1}),
      index_to_token=(1, 2, 3, 4, 5),
      token_to_index={1: 0, 2: 1, 3: 2, 4: 3, 5: 4},
    )
    >>> vocab3 = generate_vocabulary(frame, col="col3")
    >>> vocab3
    Vocabulary(
      counter=Counter({'a': 2, 'b': 1, 'c': 1, 'd': 1}),
      index_to_token=('a', 'b', 'c', 'd'),
      token_to_index={'a': 0, 'b': 1, 'c': 2, 'd': 3},
    )

    ```
    """
    counts = frame.get_column(col).value_counts(sort=True)
    return Vocabulary(Counter(dict(zip(counts.get_column(col), counts.get_column("count")))))



r"""Contain some utility functions for testing."""

from __future__ import annotations

__all__ = ["gdown_available", "requests_available", "tqdm_available"]

from arctix.testing.fixtures import gdown_available, requests_available, tqdm_available



r"""Contain utility functions to download data assets."""

from __future__ import annotations

__all__ = ["download_drive_file", "download_url_to_file"]

from pathlib import Path
from typing import TYPE_CHECKING, Any

from iden.io.utils import generate_unique_tmp_path
from iden.utils.path import sanitize_path

from arctix.utils.imports import (
    check_gdown,
    check_requests,
    is_gdown_available,
    is_requests_available,
    is_tqdm_available,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

if is_gdown_available():
    import gdown
else:  # pragma: no cover
    gdown = None

if is_requests_available():
    import requests
else:  # pragma: no cover
    requests = None

if is_tqdm_available():
    from tqdm import tqdm
else:  # pragma: no cover

    def tqdm(it: Iterable, *args: Any, **kwargs: Any) -> Iterable:  # noqa: ARG001
        return it


def download_drive_file(url: str, path: Path, *args: Any, **kwargs: Any) -> None:
    r"""Download a file from Google Drive.

    Args:
        url: The Google Drive URL.
        path: The path where to store the downloaded file.
        *args: See the documentation of ``gdown.download``.
        **kwargs: See the documentation of ``gdown.download``.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.utils.download import download_drive_file
    >>> download_drive_file(
    ...     "https://drive.google.com/open?id=1R3z_CkO1uIOhu4y2Nh0pCHjQQ2l-Ab9E",
    ...     Path("/path/to/data.tar.gz"),
    ...     quiet=False,
    ...     fuzzy=True,
    ... )  # doctest: +SKIP

    ```
    """
    check_gdown()
    path.parent.mkdir(parents=True, exist_ok=True)
    if not path.is_file():
        # Save to tmp, then commit by moving the file in case the job gets
        # interrupted while writing the file
        tmp_path = path.with_name(f"{path.name}.tmp")
        gdown.download(url, tmp_path.as_posix(), *args, **kwargs)
        tmp_path.rename(path)


def download_url_to_file(
    url: str, dst: Path | str, progress: bool = True, timeout: float = 10.0
) -> None:
    r"""Download object at the given URL to a local path.

    Args:
        url: The URL of the object to download
        dst: The path where to store the downloaded file.
        progress: If ``True``, it displays a progress bar.
        timeout: The number of second to wait until to time out.

    Example usage:

    ```pycon

    >>> from pathlib import Path
    >>> from arctix.utils.download import download_url_to_file
    >>> download_url_to_file(
    ...     'https://s3.amazonaws.com/pytorch/models/resnet18-5c106cde.pth',
    ...     Path('/tmp/temporary_file'),
    ... )  # doctest: +SKIP

    ```
    """
    check_requests()
    dst = sanitize_path(dst)
    dst.parent.mkdir(exist_ok=True, parents=True)

    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_dst = generate_unique_tmp_path(dst)

    with (
        requests.get(url, stream=True, timeout=timeout) as response,
        Path.open(tmp_dst, mode="wb") as file,
    ):
        chunks = response.iter_content(chunk_size=1024)
        if progress:
            chunks = tqdm(chunks, desc="downloading URL to file")
        for chunk in chunks:
            file.write(chunk)

    tmp_dst.rename(dst)



r"""Contain vocabulary-based transformers."""

from __future__ import annotations

__all__ = ["IndexToTokenDataFrameTransformer", "TokenToIndexDataFrameTransformer"]

from typing import TYPE_CHECKING, Any

import polars as pl

from arctix.transformer.dataframe.replace import ReplaceDataFrameTransformer

if TYPE_CHECKING:
    from arctix.utils.vocab import Vocabulary


class IndexToTokenDataFrameTransformer(ReplaceDataFrameTransformer):
    r"""Replace.

    Args:
        vocab: The vocabulary which contains the index to token
            mapping.
        index_column: The column name which contains the input indices.
        token_column: The column name which contains the output tokens.
        **kwargs: The keyword arguments to pass to ``replace``.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> import polars as pl
    >>> from arctix.transformer.dataframe import IndexToToken
    >>> from arctix.utils.vocab import Vocabulary
    >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2, "d": 4}))
    >>> vocab.get_index_to_token()
    ('b', 'a', 'c', 'd')
    >>> transformer = IndexToToken(
    ...     vocab=vocab,
    ...     index_column="col",
    ...     token_column="token",
    ... )
    >>> transformer
    IndexToTokenDataFrameTransformer(orig_column=col, final_column=token)
    >>> frame = pl.DataFrame({"col": [1, 0, 2, 3, 1]})
    >>> frame
    shape: (5, 1)
    ┌─────┐
    │ col │
    │ --- │
    │ i64 │
    ╞═════╡
    │ 1   │
    │ 0   │
    │ 2   │
    │ 3   │
    │ 1   │
    └─────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌─────┬───────┐
    │ col ┆ token │
    │ --- ┆ ---   │
    │ i64 ┆ str   │
    ╞═════╪═══════╡
    │ 1   ┆ a     │
    │ 0   ┆ b     │
    │ 2   ┆ c     │
    │ 3   ┆ d     │
    │ 1   ┆ a     │
    └─────┴───────┘

    ```
    """

    def __init__(
        self,
        vocab: Vocabulary,
        index_column: str,
        token_column: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            orig_column=index_column,
            final_column=token_column,
            old=list(range(len(vocab))),
            new=vocab.get_index_to_token(),
            **kwargs,
        )


class TokenToIndexDataFrameTransformer(ReplaceDataFrameTransformer):
    r"""Replace the values in a column by the values in a mapping.

    Args:
        vocab: The vocabulary which contains the token to index
            mapping.
        token_column: The column name which contains the input tokens.
        index_column: The column name which contains the output
            indices.
        **kwargs: The keyword arguments to pass to ``replace``.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> import polars as pl
    >>> from arctix.transformer.dataframe import TokenToIndex
    >>> from arctix.utils.vocab import Vocabulary
    >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2, "d": 4}))
    >>> vocab.get_token_to_index()
    {'b': 0, 'a': 1, 'c': 2, 'd': 3}
    >>> transformer = TokenToIndex(
    ...     vocab=vocab, token_column="col", index_column="index"
    ... )
    >>> transformer
    TokenToIndexDataFrameTransformer(orig_column=col, final_column=index)
    >>> frame = pl.DataFrame({"col": ["a", "b", "c", "d", "a"]})
    >>> frame
    shape: (5, 1)
    ┌─────┐
    │ col │
    │ --- │
    │ str │
    ╞═════╡
    │ a   │
    │ b   │
    │ c   │
    │ d   │
    │ a   │
    └─────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 2)
    ┌─────┬───────┐
    │ col ┆ index │
    │ --- ┆ ---   │
    │ str ┆ i64   │
    ╞═════╪═══════╡
    │ a   ┆ 1     │
    │ b   ┆ 0     │
    │ c   ┆ 2     │
    │ d   ┆ 3     │
    │ a   ┆ 1     │
    └─────┴───────┘

    ```
    """

    def __init__(
        self,
        vocab: Vocabulary,
        token_column: str,
        index_column: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            orig_column=token_column,
            final_column=index_column,
            old=vocab.get_token_to_index(),
            **(kwargs or {"return_dtype": pl.Int64}),
        )



r"""Contain DataFrame transformers."""

from __future__ import annotations

__all__ = [
    "BaseDataFrameTransformer",
    "Cast",
    "CastDataFrameTransformer",
    "Function",
    "FunctionDataFrameTransformer",
    "IndexToToken",
    "IndexToTokenDataFrameTransformer",
    "Replace",
    "ReplaceDataFrameTransformer",
    "Sequential",
    "SequentialDataFrameTransformer",
    "Sort",
    "SortDataFrameTransformer",
    "StripChars",
    "StripCharsDataFrameTransformer",
    "TokenToIndex",
    "TokenToIndexDataFrameTransformer",
    "is_dataframe_transformer_config",
    "setup_dataframe_transformer",
]

from arctix.transformer.dataframe.base import (
    BaseDataFrameTransformer,
    is_dataframe_transformer_config,
    setup_dataframe_transformer,
)
from arctix.transformer.dataframe.casting import CastDataFrameTransformer
from arctix.transformer.dataframe.casting import CastDataFrameTransformer as Cast
from arctix.transformer.dataframe.function import FunctionDataFrameTransformer
from arctix.transformer.dataframe.function import (
    FunctionDataFrameTransformer as Function,
)
from arctix.transformer.dataframe.replace import ReplaceDataFrameTransformer
from arctix.transformer.dataframe.replace import ReplaceDataFrameTransformer as Replace
from arctix.transformer.dataframe.sequential import SequentialDataFrameTransformer
from arctix.transformer.dataframe.sequential import (
    SequentialDataFrameTransformer as Sequential,
)
from arctix.transformer.dataframe.sorting import SortDataFrameTransformer
from arctix.transformer.dataframe.sorting import SortDataFrameTransformer as Sort
from arctix.transformer.dataframe.string import StripCharsDataFrameTransformer
from arctix.transformer.dataframe.string import (
    StripCharsDataFrameTransformer as StripChars,
)
from arctix.transformer.dataframe.vocab import IndexToTokenDataFrameTransformer
from arctix.transformer.dataframe.vocab import (
    IndexToTokenDataFrameTransformer as IndexToToken,
)
from arctix.transformer.dataframe.vocab import TokenToIndexDataFrameTransformer
from arctix.transformer.dataframe.vocab import (
    TokenToIndexDataFrameTransformer as TokenToIndex,
)



r"""Contain some utility functions to remove items in DataFrames."""

from __future__ import annotations

__all__ = ["drop_duplicates"]

import logging
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


def drop_duplicates(frame: pl.DataFrame, *args: Any, **kwargs: Any) -> pl.DataFrame:
    r"""Return DataFrame with duplicate rows removed.

    Args:
        frame: The input DataFrame.
        *args: Positional arguments that are passed to
            ``polars.DataFrame.unique``.
        **kwargs: Keyword arguments that are passed to
            ``polars.DataFrame.unique``.

    Returns:
        The DataFrame with duplicate rows removed.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.utils.dataframe import drop_duplicates
    >>> df = pl.DataFrame({"col1": [1, 2, 3, 1, 2, 3], "col2": [4, 5, 6, 4, 5, 6]})
    >>> out = drop_duplicates(df, keep="first", maintain_order=True)
    >>> out
    shape: (3, 2)
    ┌──────┬──────┐
    │ col1 ┆ col2 │
    │ ---  ┆ ---  │
    │ i64  ┆ i64  │
    ╞══════╪══════╡
    │ 1    ┆ 4    │
    │ 2    ┆ 5    │
    │ 3    ┆ 6    │
    └──────┴──────┘

    ```
    """
    orig_shape = frame.shape
    if orig_shape[0] == 0:
        return frame
    logger.info("removing duplicate rows...")
    out = frame.unique(*args, **kwargs)
    logger.info(f"removed {orig_shape[0] - out.shape[0]:,} rows | {orig_shape} -> {out.shape}")
    return out



r"""Implement some utility functions to manage optional dependencies."""

from __future__ import annotations

__all__ = [
    "check_gdown",
    "check_matplotlib",
    "check_requests",
    "check_tqdm",
    "gdown_available",
    "is_gdown_available",
    "is_matplotlib_available",
    "is_requests_available",
    "is_tqdm_available",
    "matplotlib_available",
    "requests_available",
    "tqdm_available",
]

from importlib.util import find_spec
from typing import TYPE_CHECKING, Any

from coola.utils.imports import decorator_package_available

if TYPE_CHECKING:
    from collections.abc import Callable

#################
#     gdown     #
#################


def is_gdown_available() -> bool:
    r"""Indicate if the ``gdown`` package is installed or not.

    Returns:
        ``True`` if ``gdown`` is available otherwise ``False``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import is_gdown_available
    >>> is_gdown_available()

    ```
    """
    return find_spec("gdown") is not None


def check_gdown() -> None:
    r"""Check if the ``gdown`` package is installed.

    Raises:
        RuntimeError: if the ``gdown`` package is not installed.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import check_gdown
    >>> check_gdown()

    ```
    """
    if not is_gdown_available():
        msg = (
            "`gdown` package is required but not installed. "
            "You can install `gdown` package with the command:\n\n"
            "pip install gdown\n"
        )
        raise RuntimeError(msg)


def gdown_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``gdown``
    package is installed.

    Args:
        fn: Specifies the function to execute.

    Returns:
        A wrapper around ``fn`` if ``gdown`` package is installed,
            otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import gdown_available
    >>> @gdown_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_gdown_available)


######################
#     matplotlib     #
######################


def is_matplotlib_available() -> bool:
    r"""Indicate if the ``matplotlib`` package is installed or not.

    Returns:
        ``True`` if ``matplotlib`` is available otherwise ``False``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import is_matplotlib_available
    >>> is_matplotlib_available()

    ```
    """
    return find_spec("matplotlib") is not None


def check_matplotlib() -> None:
    r"""Check if the ``matplotlib`` package is installed.

    Raises:
        RuntimeError: if the ``matplotlib`` package is not installed.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import check_matplotlib
    >>> check_matplotlib()

    ```
    """
    if not is_matplotlib_available():
        msg = (
            "`matplotlib` package is required but not installed. "
            "You can install `matplotlib` package with the command:\n\n"
            "pip install matplotlib\n"
        )
        raise RuntimeError(msg)


def matplotlib_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if
    ``matplotlib`` package is installed.

    Args:
        fn: Specifies the function to execute.

    Returns:
        A wrapper around ``fn`` if ``matplotlib`` package is installed,
            otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import matplotlib_available
    >>> @matplotlib_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_matplotlib_available)


####################
#     requests     #
####################


def is_requests_available() -> bool:
    r"""Indicate if the ``requests`` package is installed or not.

    Returns:
        ``True`` if ``requests`` is available otherwise ``False``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import is_requests_available
    >>> is_requests_available()

    ```
    """
    return find_spec("requests") is not None


def check_requests() -> None:
    r"""Check if the ``requests`` package is installed.

    Raises:
        RuntimeError: if the ``requests`` package is not installed.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import check_requests
    >>> check_requests()

    ```
    """
    if not is_requests_available():
        msg = (
            "`requests` package is required but not installed. "
            "You can install `requests` package with the command:\n\n"
            "pip install requests\n"
        )
        raise RuntimeError(msg)


def requests_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``requests``
    package is installed.

    Args:
        fn: Specifies the function to execute.

    Returns:
        A wrapper around ``fn`` if ``requests`` package is installed,
            otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import requests_available
    >>> @requests_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_requests_available)


################
#     tqdm     #
################


def is_tqdm_available() -> bool:
    r"""Indicate if the ``tqdm`` package is installed or not.

    Returns:
        ``True`` if ``tqdm`` is available otherwise ``False``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import is_tqdm_available
    >>> is_tqdm_available()

    ```
    """
    return find_spec("tqdm") is not None


def check_tqdm() -> None:
    r"""Check if the ``tqdm`` package is installed.

    Raises:
        RuntimeError: if the ``tqdm`` package is not installed.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import check_tqdm
    >>> check_tqdm()

    ```
    """
    if not is_tqdm_available():
        msg = (
            "`tqdm` package is required but not installed. "
            "You can install `tqdm` package with the command:\n\n"
            "pip install tqdm\n"
        )
        raise RuntimeError(msg)


def tqdm_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``tqdm``
    package is installed.

    Args:
        fn: Specifies the function to execute.

    Returns:
        A wrapper around ``fn`` if ``tqdm`` package is installed,
            otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arctix.utils.imports import tqdm_available
    >>> @tqdm_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_tqdm_available)



r"""Contain the vocabulary class."""

from __future__ import annotations

__all__ = ["Vocabulary"]

import logging
from collections import Counter
from collections.abc import Hashable
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from coola import objects_are_equal
from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from coola.equality import EqualityConfig

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Hashable)


class Vocabulary(Generic[T]):
    r"""Implement a vocabulary built from a counter of tokens.

    Args:
        counter: The counter used to generate the vocabulary.
            The order of the items in the counter is used to define
            the index-to-token and token-to-index mappings.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> from arctix.utils.vocab import Vocabulary
    >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
    >>> vocab.get_index("a")
    1

    ```
    """

    def __init__(self, counter: Counter) -> None:
        self._counter = counter
        self._index_to_token = tuple(self._counter.keys())
        self._token_to_index = {token: i for i, token in enumerate(self._index_to_token)}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  counter={self.counter},\n"
            f"  index_to_token={self.get_index_to_token()},\n"
            f"  token_to_index={self.get_token_to_index()},\n"
            f")"
        )

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(vocab_size={self.get_vocab_size():,})"

    def __len__(self) -> int:
        return len(self._counter)

    @property
    def counter(self) -> Counter:
        r"""The counter of the vocabulary."""
        return self._counter

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two vocabularies are equal or not.

        Args:
            other: The value to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the vocabularies are equal,
                ``False`` otherwise.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab.equal(Vocabulary(Counter({"b": 4, "a": 1, "c": 2})))
        False

        ```
        """
        if not isinstance(other, Vocabulary):
            return False
        return (
            objects_are_equal(self.counter, other.counter, equal_nan=equal_nan)
            and objects_are_equal(self.get_index_to_token(), other.get_index_to_token())
            and objects_are_equal(self.get_token_to_index(), other.get_token_to_index())
        )

    def get_index(self, token: T) -> int:
        r"""Return the index for a given index.

        Args:
            token: The token.

        Returns:
            The token index.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab.get_index("a")
        1
        >>> vocab.get_index("b")
        0
        >>> vocab.get_index("c")
        2

        ```
        """
        return self._token_to_index[token]

    def get_index_to_token(self) -> tuple[T, ...]:
        r"""Return the index to token mapping.

        Returns:
            The index to token mapping.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab.get_index_to_token()
        ('b', 'a', 'c')

        ```
        """
        return self._index_to_token

    def get_token(self, index: int) -> T:
        r"""Return the token for a given index.

        Args:
            index: The index.

        Returns:
            The token associated to the index.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab.get_token(0)
        'b'
        >>> vocab.get_token(1)
        'a'
        >>> vocab.get_token(2)
        'c'

        ```
        """
        return self._index_to_token[index]

    def get_token_to_index(self) -> dict[T, int]:
        r"""Return the token to index mapping.

        Returns:
            The token to index mapping.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab.get_token_to_index()
        {'b': 0, 'a': 1, 'c': 2}

        ```
        """
        return self._token_to_index

    def get_vocab_size(self) -> int:
        r"""Return the vocabulary size.

        Returns:
            The vocabulary size.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab.get_vocab_size()
        3

        ```
        """
        return len(self)

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Load a state dict to the current vocabulary.

        Args:
            state_dict: The state dict to load.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({}))
        >>> vocab.state_dict()
        {'counter': Counter(), 'index_to_token': (), 'token_to_index': {}}
        >>> vocab.load_state_dict(
        ...     {
        ...         "index_to_token": ("b", "a", "c"),
        ...         "token_to_index": {"b": 0, "a": 1, "c": 2},
        ...         "counter": Counter({"b": 3, "a": 1, "c": 2}),
        ...     }
        ... )
        >>> vocab.state_dict()
        {'counter': Counter({'b': 3, 'c': 2, 'a': 1}),
         'index_to_token': ('b', 'a', 'c'),
         'token_to_index': {'b': 0, 'a': 1, 'c': 2}}

        ```
        """
        self._counter = state_dict["counter"]
        self._index_to_token = state_dict["index_to_token"]
        self._token_to_index = state_dict["token_to_index"]

    def state_dict(self) -> dict:
        r"""Return the state dict of the vocabulary.

        Returns:
            The state dict which contains 3 keys: ``"counter"``,
                ``"index_to_token"``, and ``"token_to_index"``.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab.state_dict()
        {'counter': Counter({'b': 3, 'c': 2, 'a': 1}),
         'index_to_token': ('b', 'a', 'c'),
         'token_to_index': {'b': 0, 'a': 1, 'c': 2}}

        ```
        """
        return {
            "counter": self._counter,
            "index_to_token": self._index_to_token,
            "token_to_index": self._token_to_index,
        }

    def add(self, other: Vocabulary) -> Vocabulary:
        r"""Create a new vocabulary where elements from ``other`` are
        added to ``self``.

        Args:
            other: The vocabulary to add.

        Returns:
            The new vocabulary.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab1 = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab2 = Vocabulary(Counter({"b": 3, "d": 7}))
        >>> vocab = vocab1.add(vocab2)
        >>> vocab.counter
        Counter({'d': 7, 'b': 6, 'c': 2, 'a': 1})
        >>> vocab.get_index_to_token()
        ('b', 'a', 'c', 'd')

        ```
        """
        return Vocabulary(self.counter + other.counter)

    def sub(self, other: Vocabulary) -> Vocabulary:
        r"""Create a new vocabulary where elements from ``other`` are
        removed from ``self``.

        Args:
            other: The vocabulary to subtract.

        Returns:
            The new vocabulary.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab1 = Vocabulary(Counter({"b": 3, "a": 1, "c": 2}))
        >>> vocab2 = Vocabulary(Counter({"b": 3, "d": 7}))
        >>> vocab = vocab1.sub(vocab2)
        >>> vocab.counter
        Counter({'c': 2, 'a': 1})
        >>> vocab.get_index_to_token()
        ('a', 'c')

        ```
        """
        return Vocabulary(self.counter - other.counter)

    def sort_by_count(self, descending: bool = True) -> Vocabulary:
        r"""Create a new vocabulary where the counter is sorted by count.

        If multiple tokens have the same count, they are sorted by
        token values.

        Args:
            descending: If ``True``, the items are sorted in
                descending order by token.

        Returns:
            The new vocabulary where the counter is sorted by count.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2})).sort_by_count()
        >>> vocab.counter
        Counter({'b': 3, 'c': 2, 'a': 1})
        >>> vocab.get_index_to_token()
        ('b', 'c', 'a')

        ```
        """
        return Vocabulary(
            Counter(
                dict(
                    sorted(
                        self.counter.items(),
                        key=lambda item: (item[1], item[0]),
                        reverse=descending,
                    )
                )
            )
        )

    def sort_by_token(self, descending: bool = False) -> Vocabulary:
        r"""Create a new vocabulary where the counter is sorted by token.

        Args:
            descending: If ``True``, the items are sorted in
                descending order by token.

        Returns:
            The new vocabulary where the counter is sorted by token.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2})).sort_by_token()
        >>> vocab.counter
        Counter({'b': 3, 'c': 2, 'a': 1})
        >>> vocab.get_index_to_token()
        ('a', 'b', 'c')

        ```
        """
        return Vocabulary(Counter(dict(sorted(self.counter.items(), reverse=descending))))

    def most_common(self, max_num_tokens: int) -> Vocabulary:
        r"""Get a new vocabulary with the ``max_num_tokens`` most common
        tokens of the current vocabulary.

        Args:
            max_num_tokens: The maximum number of tokens.

        Returns:
            The new vocabulary with the most common tokens.
                The counter is sorted by decreasing order of count.

        Example usage:

        ```pycon

        >>> from collections import Counter
        >>> from arctix.utils.vocab import Vocabulary
        >>> vocab = Vocabulary(Counter({"b": 3, "a": 1, "c": 2})).most_common(2)
        >>> vocab.counter
        Counter({'b': 3, 'c': 2})
        >>> vocab.get_index_to_token()
        ('b', 'c')

        ```
        """
        return Vocabulary(Counter(dict(self.counter.most_common(max_num_tokens))))


class VocabularyEqualityComparator(BaseEqualityComparator[Vocabulary]):
    r"""Implement an equality comparator for ``Vocabulary`` objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> VocabularyEqualityComparator:
        return self.__class__()

    def equal(self, actual: Vocabulary, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(Vocabulary):  # pragma: no cover
    EqualityTester.add_comparator(Vocabulary, VocabularyEqualityComparator())



r"""Contain ``polars.DataFrame`` transformers to process string
values."""

from __future__ import annotations

__all__ = ["StripCharsDataFrameTransformer"]

from typing import TYPE_CHECKING, Any

import polars as pl

from arctix.transformer.dataframe.base import BaseDataFrameTransformer
from arctix.utils.imports import is_tqdm_available

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

if is_tqdm_available():
    from tqdm import tqdm
else:  # pragma: no cover

    def tqdm(it: Iterable, *args: Any, **kwargs: Any) -> Iterable:  # noqa: ARG001
        return it


class StripCharsDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implement a transformer to remove leading and trailing
    characters.

    Args:
        columns: The columns to prepare.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arctix.transformer.dataframe import StripChars
    >>> transformer = StripChars(columns=["col2", "col3"])
    >>> transformer
    StripCharsDataFrameTransformer(columns=('col2', 'col3'))
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬───────┬───────┐
    │ col1 ┆ col2 ┆ col3  ┆ col4  │
    │ ---  ┆ ---  ┆ ---   ┆ ---   │
    │ i64  ┆ str  ┆ str   ┆ str   │
    ╞══════╪══════╪═══════╪═══════╡
    │ 1    ┆ 1    ┆ a     ┆ a     │
    │ 2    ┆ 2    ┆  b    ┆  b    │
    │ 3    ┆ 3    ┆   c   ┆   c   │
    │ 4    ┆ 4    ┆ d     ┆ d     │
    │ 5    ┆ 5    ┆ e     ┆ e     │
    └──────┴──────┴───────┴───────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬───────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4  │
    │ ---  ┆ ---  ┆ ---  ┆ ---   │
    │ i64  ┆ str  ┆ str  ┆ str   │
    ╞══════╪══════╪══════╪═══════╡
    │ 1    ┆ 1    ┆ a    ┆ a     │
    │ 2    ┆ 2    ┆ b    ┆  b    │
    │ 3    ┆ 3    ┆ c    ┆   c   │
    │ 4    ┆ 4    ┆ d    ┆ d     │
    │ 5    ┆ 5    ┆ e    ┆ e     │
    └──────┴──────┴──────┴───────┘

    ```
    """

    def __init__(self, columns: Sequence[str]) -> None:
        self._columns = tuple(columns)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(columns={self._columns})"

    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        for col in tqdm(self._columns, desc="stripping chars"):
            frame = frame.with_columns(frame.select(pl.col(col).str.strip_chars()))
        return frame
